import numpy as np
from syndalib.makers.circle_maker import CircleMaker
from syndalib.makers.ellipse_maker import EllipseMaker
from syndalib.makers.line_maker import LineMaker
from syndalib.makers.maker import Maker


class ConicMaker(Maker):
    """
    Maker implementation for conics sampled from a 2D space [x_min,x_max] x [y_min, y_max]
    you can define the 2D space by calling the function syndalib.syn2d(...).
    A conic so far is randomly sampled among circles, lines and ellipses.
    """
    def __init__(self):
        Maker.MODEL = "conics"
        self.circle_maker = CircleMaker()
        self.line_maker = LineMaker()
        self.ellipse_maker = EllipseMaker()

        super(ConicMaker, self).__init__()

    def generate_random_model(self, n_inliers):
        rand_num = np.random.randint(0, 3)

        if rand_num == 0:
            return self.circle_maker.generate_random_model(n_inliers)
        if rand_num == 1:
            return self.line_maker.generate_random_model(n_inliers)
        if rand_num == 2:
            return self.ellipse_maker.generate_random_model(n_inliers)

    def point(self):
        pass



import string
import os
from typing import Type, List

from syndalib.makers.circle_maker import CircleMaker
from syndalib.makers.conic_maker import ConicMaker
from syndalib.makers.ellipse_maker import EllipseMaker
from syndalib.makers.line_maker import LineMaker
from syndalib.makers.maker import Maker
from syndalib.utils.config import opts


def set_options(options: dict) -> None:
    """
    specify how to generate outliers and samples of specified classes.\n
    CLASS_TYPES = "outliers","circles", "lines", "ellipses", "conics".\n
    CLASS_OPTIONS are the values of the options dictionary and they vary with respect to the corresponding key.\n
    CLASS_OPTIONS always consists of a dictionary whose values are tuples of 2 floats
    (it represents the interval of values the specific parameter can take).\n
    the default options are the following:\n
    options = {
        "outliers": {
                    "x_r": (-2.5, 2.5),\n
                    "y_r": (-2.5, 2.5)
        },\n
        "circles": {
         \t   "radius_r": (0.5, 1.5),\n
         \t   "x_center_r": (-1.0, 1.0),\n
         \t   "y_center_r": (-1.0, 1.0),
        },\n
        "lines": {
                    "x_r": (-2.5, 2.5),\n
                    "y_r": (-2.5, 2.5)
        },\n
        "ellipses": {
            "radius_r": (0.5, 1.5),\n
            "x_center_r": (-1, 1),\n
            "y_center_r": (-1, 1),\n
            "width_r": (0.1, 1),\n
            "height_r": (0.1, 1)
        },\n
    }


    :param options: dict, a dictionary in the form {"CLASS_TYPE": CLASS_OPTIONS, ... }, where class_options is a class specific dictionary.
    :return:
    """

    for key in options.keys():
        opts[key] = options[key]


def generate_data(
    ns: int,
    npps: int,
    class_type: str,
    nm: int,
    outliers_range: List,
    noise_range: List,
    ds_name: str,
    is_train: bool = True,
    format: str = "matlab"):
    """
    generates .mat files in a structured fashion within the specified folder

    :param ns: int, total number of samples to be generated for each outlier rate
    :param npps: int, total number of points per sample
    :param class_type: str, models to be created. the possible models are: "circles","lines","ellipses","conics".
    :param nm: int, the number of models to be randomly sampled for each sample in the dataset, the default number is 1.
    :param outliers_range: list of floats, list of outliers rates (ex. outliers_range = [0.10,0.20] will generate samples with 10% and 20% outliers percentage
    :param noise_range: list of floats, list of stddevs of the gaussian noise to be added to the inliers (ex. noise range = [0.01, 0.02, 0.03] will generate samples with gaussian noise (0,0.01), (0,0.02) and (0, 0.03).
    :param ds_name: str, format of file "matlab" is the only one supported so far. default is "matlab".
    :param is_train: bool, true to write data in train folder, false to write data in test folder. default is True.

    :param format:
    :return:
    """

    Maker.NUM_SAMPLES = ns
    Maker.NUM_POINTS_PER_SAMPLE = npps
    Maker.OUTLIERS_PERC_RANGE = outliers_range
    Maker.MODEL = class_type
    Maker.NOISE_PERC_RANGE = noise_range
    Maker.DEST = format
    if is_train:
        Maker.TRAIN_DIR = "train"
        Maker.TEST_DIR = ""
    else:
        Maker.TRAIN_DIR = ""
        Maker.TEST_DIR = "test"

    Maker.NUMBER_OF_MODELS = nm

    # create basedir
    basedir = class_type
    if not os.path.exists(basedir):
        os.mkdir(basedir)

    basedir += "/nm_" + str(nm)
    if not (os.path.exists(basedir)):
        os.mkdir(basedir)

    basedir += "/" + ds_name
    if not os.path.exists(basedir):
        os.mkdir(basedir)

    basedir += "/npps_" + str(npps)
    if not os.path.exists(basedir):
        os.mkdir(basedir)

    basedir += "/ns_" + str(ns)
    if not os.path.exists(basedir):
        os.mkdir(basedir)

    Maker.BASE_DIR = basedir

    # create img dirs
    imgbasedir = Maker.BASE_DIR
    if Maker.TRAIN_DIR != "":
        imgbasedir += "/" + "train"
    if Maker.TEST_DIR != "":
        imgbasedir += "/" + "test"

    if not os.path.exists(imgbasedir):
        os.mkdir(imgbasedir)

    imgbasedir += "/" + "imgs"
    if not os.path.exists(imgbasedir):
        os.mkdir(imgbasedir)

    Maker.IMG_BASE_DIR = imgbasedir

    maker = None
    if Maker.MODEL == "circles":
        maker = CircleMaker()
    elif Maker.MODEL == "lines":
        maker = LineMaker()
    elif Maker.MODEL == "ellipses":
        maker = EllipseMaker()
    elif Maker.MODEL == "conics":
        maker = ConicMaker()
    if maker is not None:
        maker.start()
    else:
        print('Incorrect name for "model"')



import math
import random
import numpy as np

from typing import Tuple


def circle_points(r: float,
                  c: Tuple[float, float],
                  n: int):
    """

    :param r: radius
    :param c: center (x,y)
    :param n: number of points
    :return: np array [(x_1,y_1,1),...,(x_np,y_np,1)]
    """
    points = []
    for _ in range(n):
        alpha = 2 * math.pi * random.random()
        x = c[0] + r * math.cos(alpha)
        y = c[1] + r * math.sin(alpha)
        points.append((x, y, 1))

    return np.array(points)


def line_points(point_1: Tuple[float, float],
                point_2: Tuple[float, float],
                n: int):
    """
    samples point from a line obtained by connecting two points
    :param point_1: (float, float)
    :param point_2: (float, float)
    :param n: number of points to be drawn
    :return: np.array, [(x_1,y_1,1), ..., (x_n,y_n,1)]
    """
    x1 = point_1[0]
    y1 = point_1[1]
    x2 = point_2[0]
    y2 = point_2[1]
    m = (y1-y2)/(x1-x2)
    q = y1-m*x1
    points = []
    for _ in range(n):
        x = random.uniform(point_1[0], point_2[0])
        y = m*x + q
        points.append((x, y, 1))
    return np.array(points)



import abc
import math
import os
from abc import ABC
from random import uniform
from typing import Tuple, List

import numpy as np
import scipy.io
from syndalib.utils.config import opts
from syndalib.utils.utils import (
    compute_num_inliers_per_model,
    plot_sample,
    convert_to_np_struct,
    convert_to_mat_struct,
)


class Maker(ABC):
    """
    abstract class. serves as reference to create all subclasses that
    will implement the generation of models of specific classes.
    """
    NUM_SAMPLES = None
    NUM_POINTS_PER_SAMPLE = None
    BASE_DIR = None
    TRAIN_DIR = None
    TEST_DIR = None
    CUR_DIR = None
    NUMBER_OF_MODELS = None
    OUTLIERS_PERC_RANGE = None
    NOISE_PERC_RANGE = None
    NOISE_PERC = None
    OUTLIERS_PERC = None
    INLIERS_RANGE = None
    IMG_BASE_DIR = None
    IMG_DIR = None
    MODEL = None
    DT = np.dtype([("x1p", "O"), ("x2p", "O"), ("labels", "O")])

    @abc.abstractmethod
    def generate_random_model(self, n_inliers: int) -> List:
        """
        generate a random model. the generating process changes for each class.

        :param n_inliers: int, number of inliers to be generated
        :return: list, list of tuples (2d coordinates) of inliers
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def point(*args, **kwargs) -> Tuple:
        """
        generates a point of the random model of the specified class

        :param args: any parameter needed to generate model
        :param kwargs:
        :return: a tuple (x,y) representing a 2D point
        """
        pass

    def generate_random_sample(self, plot: bool = True) -> Tuple:
        """
        generates a random sample of the desired class.

        :param: bool, true to plot sample, false otherwise. default is True.
        :return: tuple, (x1,x2,labels) ---
        where:
        x1 is an np.ndarray (npps,);
        x2 is an np.ndarray (npps,);
        labels is an np.ndarray (npps, nm);
        """
        tot_inliers = np.random.choice(Maker.INLIERS_RANGE)
        inliers_per_model = compute_num_inliers_per_model(
            tot_inliers, Maker.NUMBER_OF_MODELS
        )
        inliers = []
        for n_inliers in inliers_per_model:
            # model_inliers is a list of tuples (x,y)
            model_inliers = self.generate_random_model(n_inliers=n_inliers)
            inliers.append(model_inliers)

        n_outliers = Maker.NUM_POINTS_PER_SAMPLE - tot_inliers
        outliers = self.generate_outliers(n_outliers)

        if plot:
            plot_sample(inliers, outliers, Maker.IMG_DIR)

        sample = convert_to_np_struct(inliers=inliers, outliers=outliers)
        np.random.shuffle(sample)
        sample_mat = convert_to_mat_struct(sample)

        return sample_mat

    def generate_dataset_fixed_nr_and_or(self) -> None:
        """
        saves a .mat file in a structured fashioned folder with a fixed gaussian noise and a fixed outliers rate

        :return:
        """

        avg_num_inliers = math.floor(
            Maker.NUM_POINTS_PER_SAMPLE * (1.0 - Maker.OUTLIERS_PERC)
        )
        Maker.INLIERS_RANGE = list(range(avg_num_inliers - 2, avg_num_inliers + 2))

        # handle possibility that training or testing directory are empty
        for curDir in [Maker.TRAIN_DIR, Maker.TEST_DIR]:
            if curDir == "":
                continue
            Maker.CUR_DIR = curDir
            Maker.IMG_DIR = (
                Maker.IMG_BASE_DIR
                + "/"
                + Maker.MODEL
                + "_no_"
                + str(int(Maker.OUTLIERS_PERC * 100))
                + "_noise_"
                + str(Maker.NOISE_PERC)
            )
            if not os.path.exists(Maker.IMG_DIR):
                os.mkdir(Maker.IMG_DIR)

            samples = []
            for _ in range(Maker.NUM_SAMPLES):
                sample = self.generate_random_sample()
                samples.append(sample)

            samples = np.array(samples, dtype=Maker.DT)
            dataset = np.array([samples])
            # save it into a matlab file
            folder = (
                "./"
                + Maker.BASE_DIR
                + "/"
                + curDir
                + "/"
                + Maker.MODEL
                + "_no_"
                + str(int(Maker.OUTLIERS_PERC * 100))
                + "_noise_"
                + str(Maker.NOISE_PERC)
                + ".mat"
            )
            scipy.io.savemat(
                folder, mdict={"dataset": dataset, "outlierRate": Maker.OUTLIERS_PERC}
            )

    def generate_outliers(self, n_outliers: int) -> List:
        """
        generates a list of randomly sampled outliers

        :param n_outliers: int, number of outliers to be generated
        :return: list of tuples, list of randomly sampled outliers
        """
        x_min, x_max = opts["outliers"]["x_r"]
        y_min, y_max = opts["outliers"]["y_r"]

        outliers = [
            (uniform(x_min, x_max), uniform(y_min, y_max)) for _ in range(n_outliers)
        ]
        return outliers

    def start(self):
        for noise_stddev in Maker.NOISE_PERC_RANGE:
            for out_rate in Maker.OUTLIERS_PERC_RANGE:
                Maker.NOISE_PERC = noise_stddev
                Maker.OUTLIERS_PERC = out_rate
                self.generate_dataset_fixed_nr_and_or()



from random import uniform
from typing import List

from syndalib.makers.maker import Maker
import numpy as np
from syndalib.utils.config import opts


class LineMaker(Maker):
    """
    Maker implementation for lines sampled from a 2D space [x_min,x_max] x [y_min, y_max]
    you can define the 2D space by calling the function syndalib.syn2d(...)
    """
    def __init__(self):
        Maker.MODEL = "lines"
        super(LineMaker, self).__init__()

    @staticmethod
    def point(m, q, noise_perc, x_min, x_max, y_min, y_max):
        while True:
            x = uniform(x_min, x_max)
            y = m * x + q + np.random.normal(0, noise_perc)
            if y_min <= y <= y_max:
                break

        return x, y

    def generate_random_model(self, n_inliers: int) -> List:
        x_min, x_max = opts["lines"]["x_r"]
        y_min, y_max = opts["lines"]["y_r"]

        x1 = uniform(x_min, x_max)
        y1 = uniform(y_min, y_max)

        x2 = uniform(x_min, x_max)
        y2 = uniform(y_min, y_max)

        m = (y2 - y1) / (x2 - x1)
        q = y1 - m*x1

        xy = [LineMaker.point(m, q, LineMaker.NOISE_PERC, x_min, x_max, y_min, y_max) for _ in range(n_inliers)]
        return xy



import string
from math import floor
from typing import List, Tuple

import matplotlib.pyplot as plt
import random as rand
import os
import numpy as np


def getRanges(nSamplesPerOR: int):
    """

    :param nSamplesPerOR:
    :return:
    """
    xr = 1
    yr = 1
    rr = 1
    turn = 0
    while xr * yr * rr < nSamplesPerOR:
        if turn == 0:
            xr *= 2
        elif turn == 1:
            yr *= 2
        elif turn == 2:
            rr *= 2
            turn = -1
        turn += 1

    return [xr, yr, rr]


def checkExistenceAndCreate(path: string):
    """

    :param path: takes full path of mat file, included the name of the file
    :return:
    """
    folders = path.split("/")
    folders = folders[1:-1]  # dont' care about the '.' nor the name of the file
    path = folders[0]
    if not os.path.exists(path):
        os.mkdir(path)
    path = path + "/" + folders[1]
    if not os.path.exists(path):
        os.mkdir(path)
    path = path + "/" + folders[2]
    if not os.path.exists(path):
        os.mkdir(path)
    path = path + "/" + folders[3]
    if not os.path.exists(path):
        os.mkdir(path)



def convert_to_np_struct(inliers: List, outliers: List) -> np.ndarray:
    """
    merges list of inliers and list of outliers into an np.ndarray and returns it

    :param inliers: list, a list of lists of inliers. one list for each model. each list is made of tuples (coord1,coord2,...)
    :param outliers: list, a list of tuples (coord1,coord2,...)
    :return: np.ndarray, shuffled sample
    """

    # convert inliers to np array
    inliers = np.array([np.array(model_inliers) for model_inliers in inliers])

    # obtain number of models
    num_models = inliers.shape[0]

    # count all inliers by summing inliers in each model
    num_inliers = 0
    for i_model in range(num_models):
        num_inliers += inliers[i_model].shape[0]

    # obtain number of coordinates (in the circle case it is: 2)
    n_coordinates = inliers[0].shape[1]

    # create an array of ones that I'll modify step by step
    # it is an array of ones because I'll have that some point will be inlier for a model but outliers (ones) for others
    inliers_out = np.ones((num_inliers, n_coordinates + num_models))
    for i_model in range(num_models):
        # find start index row for this model
        start_i = 0
        for i in range(i_model):
            start_i += inliers[i].shape[0]
        num_inliers_in_model = inliers[i_model].shape[0]
        # find end index row for this model
        end_i = start_i + num_inliers_in_model
        inliers_out[start_i:end_i, 0:n_coordinates] = inliers[i_model]
        inliers_out[start_i:end_i, n_coordinates + i_model] = np.zeros(
            shape=(num_inliers_in_model,)
        )

    # set value to 1 for outliers
    outliers_out = np.ones((len(outliers), n_coordinates + num_models))
    outliers_out[:, 0:n_coordinates] = outliers

    # concatenate arrays and shuffle
    sample = np.concatenate((inliers_out, outliers_out), axis=0)

    return sample


def convert_to_mat_struct(sample: np.ndarray) -> Tuple:
    """
    converts np.ndarray into triplet (x1p, x2p, labels)

    :param sample: np.ndarray, sample coded as 2d points + their label
    :return: a triplet made of correspondences + labels, the kind of struct matlab deals with
    """
    sample = sample.T
    x1p = np.array([elem for elem in sample[0, :]], dtype=np.float32)
    x2p = np.array([elem for elem in sample[1, :]], dtype=np.float32)
    num_models = sample.shape[0] - 2

    num_points_per_sample = sample.shape[1]
    labels = np.zeros(shape=(num_points_per_sample, num_models))
    for i_model in range(num_models):
        labels[:, i_model] = np.array(
            [elem for elem in sample[2 + i_model, :]], dtype=np
        )
    # se dovessi avere problemi con il single model fitting forse mi basta deglutire la seconda dimensione di labels
    # oltre al commento sopra dovrei evitare di fare la seguente transpose immagino
    labels = labels.T
    circle_mat = (x1p, x2p, labels)
    return circle_mat


def compute_num_inliers_per_model(tot_num_inliers, num_of_models):
    """

    :param tot_num_inliers: total number of inliers in the whole sample
    :param num_of_models: the number of models to be drawn in each sample
    :return: a list that contains the number of inlier points for each model
    """
    n_inliers = []
    remaining = tot_num_inliers
    remaining_n_of_models = num_of_models
    while remaining_n_of_models > 0:
        cur_model_n_nliers = floor(remaining / remaining_n_of_models)
        n_inliers.append(cur_model_n_nliers)
        remaining -= cur_model_n_nliers
        remaining_n_of_models -= 1
    return n_inliers


def plot_sample(inliers, outliers, imgdir):
    for model in inliers:
        plt.scatter(*zip(*model), s=2)
    plt.scatter(*zip(*outliers), s=2)
    plt.grid(color="k", linestyle=":", linewidth=1)
    plt.axes().set_aspect("equal", "datalim")
    plt.gca().set_aspect("equal", adjustable="box")

    # save plot once in a while
    random_number = rand.randrange(10000)
    if random_number % 20 == 0:
        folder = imgdir + "/" + str(rand.randrange(10000)) + ".png"
        # checkExistenceAndCreate(folder)
        plt.savefig(folder)

    # show it
    plt.show()



from setuptools import find_packages, setup

setup(
    name="syndalib",
    packages=find_packages(
        include=[
            "syndalib",
            "syndalib.utils",
            "syndalib.makers",
        ]
    ),
    version="0.1.1",
    description="synthetic point cloud library",
    author="William Bonvini",
    license="MIT",
    install_requires=["pytest-runner"],
    tests_require=["pytest==4.4.1"],
    test_suite="tests",
)



import math
from random import uniform
import numpy as np
import random as rand

from syndalib.makers.maker import Maker
from syndalib.utils.config import opts


class EllipseMaker(Maker):
    """
    Maker implementation for ellipses sampled from a 2D space.
    you can define the 2D space by calling the function syndalib.syn2d(...).
    """

    def __init__(self):
        Maker.MODEL = "ellipses"
        super(EllipseMaker, self).__init__()

    def generate_random_model(self, n_inliers):
        x_c_min, x_c_max = opts["ellipses"]["x_center_r"]
        y_c_min, y_c_max = opts["ellipses"]["y_center_r"]
        r_min, r_max = opts["ellipses"]["radius_r"]
        w_min, w_max = opts["ellipses"]["width_r"]
        h_min, h_max = opts["ellipses"]["height_r"]

        x_center = uniform(x_c_min, x_c_max)
        y_center = uniform(y_c_min, y_c_max)
        radius = uniform(r_min, r_max)
        width = uniform(w_min, w_max)
        height = uniform(h_min, h_max)
        xy = [
            EllipseMaker.point(
                y_center, x_center, radius, width, height, EllipseMaker.NOISE_PERC
            )
            for _ in range(n_inliers)
        ]
        return xy

    @staticmethod
    def point(h, k, r, width, height, noiseperc):
        """
        draws a random point belonging to an ellipse by scaling a circle with center (h,k) and radius r

        :param h:
        :param k:
        :param r:
        :param width:
        :param height:
        :param noiseperc:
        :return:
        """
        theta = rand.random() * 2 * math.pi
        x = h + math.cos(theta) * r
        noise = np.random.normal(0, noiseperc)
        x = x + noise
        y = k + math.sin(theta) * r
        noise = np.random.normal(0, noiseperc)
        y = y + noise

        return x * width, y * height



from typing import Type
import matplotlib.pyplot as plt
import numpy as np


def plot_2d(points: Type[np.array]):
    """
    plots in 2D some points
    :param points: np.array, [(x_1,y_1,1), ..., (x_n,y_n,1)]
    :return:
    """
    xs = [point[0] for point in points]
    ys = [point[1] for point in points]
    plt.scatter(xs, ys, s=10)



import math
import os
from random import uniform
import numpy as np
import random as rand
from syndalib.makers.maker import Maker
from syndalib.utils.config import opts


class CircleMaker(Maker):
    """
    Maker implementation for circles sampled from a 2D space.
    you can define the 2D space by calling the function syndalib.syn2d(...).
    """
    def __init__(self):
        Maker.MODEL = "circles"
        super(CircleMaker, self).__init__()

    def generate_random_model(self, n_inliers: int = None):
        """
        generates points belonging to a randomly sampled model

        :param n_inliers: number of inlier points for this specific model
        :return:
        """
        x_c_min, x_c_max = opts["circles"]["x_center_r"]
        y_c_min, y_c_max = opts["circles"]["y_center_r"]
        r_min, r_max = opts["circles"]["radius_r"]

        x_center = uniform(x_c_min, x_c_max)
        y_center = uniform(y_c_min, y_c_max)
        radius = uniform(r_min, r_max)
        xy = [
            CircleMaker.point(y_center, x_center, radius, CircleMaker.NOISE_PERC)
            for _ in range(n_inliers)
        ]
        return xy

    @staticmethod
    def point(h: float, k: float, r: float, noise: float):
        """
        draws a random point belonging to a circle with center (h,k) and radius r

        :param h: x center of circle
        :param k: y center of circle
        :param r: radius of circle
        :param noise: stddev of gaussian noise
        :return:
        """

        theta = rand.random() * 2 * math.pi
        x = h + math.cos(theta) * r
        noise_x = np.random.normal(0, noise)
        x = x + noise_x
        y = k + math.sin(theta) * r
        noise_y = np.random.normal(0, noise)
        y = y + noise_y
        return x, y
